#include <stdio.h>
//#define DEBUG	1
#include "MemoryMgr_private.h"
#include "DataMgr.h"
#include "ExgMgr.h"
#include <stdlib.h>

typedef struct
{
	char * data;
	UInt32 datasize;
	UInt32 used;
} dbread;

Err DBReadFunc(void * dest, UInt32 * size, void * userdata)
{
	dbread * mydata = (dbread *)userdata;
	UInt32 realsize = (*size < (mydata->datasize - mydata->used) ? *size : (mydata->datasize - mydata->used));
	printf("Size requested: %d, size copied: %d\n", *size, realsize);
	MemMove(dest, mydata->data+mydata->used, realsize);
	mydata->used += realsize;
	*size = realsize;
	return errNone;
}

int main(int argc, char ** argv)
{
	int i;

	free_mem_node memorynode1;
	free_mem_node memorynode2;
	free_mem_node memorynode3;
	void * memorychunk1;
	void * memorychunk2;
	void * memorychunk3;
	void * someptr;
	DmOpenRef dbref;
	LocalID	dbid;
	UInt16 index;
	FILE * pdbfile;
	UInt32 filesize;
	dbread userdata;
	Boolean needreset;
	
	char data[8][25] = {"Cow says MOOOOO!\n", "Live long and prosper.\n", 
		"42 is the answer.\n", "Got Milk?\n", "Meow.\n", "Flatfoot Rocks!\n",
		"Pet the kitty.\n", "Mikie loves Rini.\n"};
	
	puts("GHOST Memory Subsystem Test\n");
	puts("Allocating initial chunks...");
	
	puts("Allocating 512 bytes.");
	memorynode1.start = (void *)malloc(512);
	memorychunk1 = memorynode1.start;
	memorynode1.size = 512;
	memorynode1.next = &memorynode2;
	memorynode1.last = NULL;
	printf("Address: %X\n", (int)memorynode1.start);
	
	puts("Allocating 1024 bytes.");
	memorynode2.start = (void *)malloc(1024);
	memorychunk2 = memorynode2.start;
	memorynode2.size = 1024;
	memorynode2.next = &memorynode3;
	memorynode2.last = &memorynode1;
	printf("Adddress: %X\n", (int)memorynode2.start);
	
	puts("Allocatibng 128KB bytes.");
	memorynode3.start = (void *)malloc(128*1024);
	memorychunk3 = memorynode3.start;
	memorynode3.size = 128*1024;
	memorynode3.next = NULL;
	memorynode3.last = &memorynode2;
	printf("Adddress: %X\n", (int)memorynode3.start);
	
	puts("Done.\n");
	
	puts("Calling initmem...");
	initmem(&memorynode1);
	puts("Done.\n");
	
	puts("Creating database \"Stuff\".");
	DmCreateDatabase(0, "Stuff", 'MoDC', 'MoDT', false);
	
	puts("Finding by type and creator...");
	dbref = DmOpenDatabaseByTypeCreator('MoDT', 'MoDC', 0);
	printf("dbref: %d\n", (int)dbref);
	
	puts("Finding by name...");
	dbid = DmFindDatabase(0, "Stuff");
	dbref = DmOpenDatabase(0, dbid, 0);
	printf("dbref: %d\n", (int)dbref);
	
	for(i = 0; i < 8; ++i)
	{
		index = dmMaxRecordIndex;
		printf("Adding new record with text: %s", data[i]);
		someptr = DmNewRecord(dbref, &index, strlen(data[i])+1);
		printf("Record added at index: %d\n", (int)index);
		fflush(stdin);
		DmWrite(someptr, 0, data[i], strlen(data[i])+1);
	}
	
	
	for(i = 0; i < 128; ++i)
	{
		index = rand()%8;
		printf("Record at index %d has text: %s", (int)index, DmQueryRecord(dbref, index));
	}
	
	puts("Creating resource database \"StuffSource\"");
	DmCreateDatabase(0, "StuffSource", 'MoDC', 'rsrc', true);
	dbref = DmOpenDatabaseByTypeCreator('rsrc', 'MoDC', 0);
	for(i = 0; i < 8; ++i)
	{
		printf("Adding new resource with type; TStr, rsrcid: %d and text: %s", i, data[i]);
		someptr = DmNewResource(dbref, 'TStr', i, strlen(data[i])+1);
		DmWrite(someptr, 0, data[i], strlen(data[i])+1);
	}
	
	for(i = 0; i < 128; ++i)
	{
		index = rand()%8;
		printf("Record with type TStr and resource id %d has text %s\n", index, DmGetResource('TStr', index));
		//printf("Record with type TStr and resource id %d has text %s\n", index, DmGetResourceIndex(dbref, index));
	}
	
	if(argc > 2)
	{
		printf("There are now %d databases.\n", DmNumDatabasesEx());
		printf("Attempting to open %s\n", argv[1]);
		pdbfile = fopen(argv[1], "rb");
		if(pdbfile)
		{
			puts("File opened successfully.");
			fseek(pdbfile, 0, SEEK_END);
			filesize = ftell(pdbfile);
			fseek(pdbfile, 0, SEEK_SET);
			
			userdata.data = malloc(filesize);
			fread(userdata.data, 1, filesize, pdbfile);
			userdata.datasize = filesize;
			userdata.used = 0;
			
			ExgDBRead(DBReadFunc, NULL, &userdata, 0, &needreset, false);
			
			printf("There are now %d databases.\n", DmNumDatabasesEx());
			
			printf("Attempting to open database %s\n", argv[2]);
			dbid = DmFindDatabase(0, argv[2]);
			if(dbid)
			{
				dbref = DmOpenDatabase(0, dbid, dmModeReadOnly);
				printf("Database loaded from file has %d records.\n", DmNumRecords(dbref));
			}
		}
	}
			
		
	
	free(memorychunk1);
	free(memorychunk2);
	free(memorychunk3);
	return 0;
	puts("Test Complete.");
}